/*
 *
 * Copyright (C) 2003 Sadysta ( http://sadysta.ath.cx )
 * Copyright (C) 2007 Tomasz Ostrowski ( http://ostry.w.szu.pl ) - modified from 1.8.9 source
 *
 * This file is part of Free GG.
 *
 * Free GG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Free GG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Free GG; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "GaduThread.h"
#pragma package(smart_init)

#include "Main.h"
#include "Talk.h"
#include "Settings.h"
#include "StatusDesc.h"
#include "Search.h"
#include "Me.h"
#include "Contact.h"
#include "Activity.h"
#include "Popup.h"
#include "Transfer.h"
#include "TransferThread.h"

#include <MMSystem.h>
#include <stdio.h>
#include <filectrl.hpp>
#include <time.h>
//---------------------------------------------------------------------------
__fastcall TGaduThread::TGaduThread(bool CreateSuspended)
        : TThread(CreateSuspended)
{
        WriteQueue = new TStringList();
}
//---------------------------------------------------------------------------
__fastcall TGaduThread::~TGaduThread()
{
        delete WriteQueue;
}
//---------------------------------------------------------------------------
AnsiString log_msg, message, status_msg;
int sender, msg_time;
//---------------------------------------------------------------------------
void __fastcall TGaduThread::LogMsg()
{
        MainForm -> Log -> Lines -> Add (log_msg);
}
//---------------------------------------------------------------------------

void __fastcall TGaduThread::SetStatusDesc()
{
        if(StatusDescForm)
            StatusDescForm->Edit1->Text = asGaduStatus;
        MainForm->TrayIcon1->Hint = "Free GG:\n" + asGaduStatus; 
}


//---------------------------------------------------------------------------
void __fastcall TGaduThread::LogMessage (AnsiString S)
{
        if(!SettingsForm->CheckBox5) return;	//zapobiega AV gdy okna s zamykane
		if (!SettingsForm -> CheckBox5 -> Checked) return;
        log_msg = S;
        Synchronize (&LogMsg);
}
//---------------------------------------------------------------------------
char *gg_urldecode (const char *str)
{
        char *buf;
        int size = 0;

        for (unsigned int i = 0; i < strlen(str); i++)
                if (str[i] == '%') size--; else size++;

        buf = (char*)malloc (size+1);

        size = 0;
        for (unsigned int i = 0; i < strlen(str); i++)
        {
                if (str[i] == '%')
                        if (str[i+1] == '%')
                        {
                                buf[size] = '%';
                                i++;
                        } else
                        {
                                int code = StrToInt ("0x" + str[i+1] + str[i+2]);
                                buf[size] = code;
                                i += 2;
                        }
                else buf[size] = str[i];

                size++;
        }
        buf[size] = 0;

        return buf;
}
//---------------------------------------------------------------------------
char *gg_urlencode(const char *str)
{
	char *q, *buf, hex[] = "0123456789abcdef";
	const char *p;
	int size = 0;

	if (!str)
		str = strdup("");

	for (p = str; *p; p++, size++) {
		if (!((*p >= 'a' && *p <= 'z') || (*p >= 'A' && *p <= 'Z') || (*p >= '0' && *p <= '9') || *p == ' '))
			size += 2;
	}

	if (!(buf = (char*)malloc(size + 1)))
		return NULL;

	for (p = str, q = buf; *p; p++, q++) {
		if ((*p >= 'a' && *p <= 'z') || (*p >= 'A' && *p <= 'Z') || (*p >= '0' && *p <= '9'))
			*q = *p;
		else {
			if (*p == ' ')
				*q = '+';
			else {
				*q++ = '%';
				*q++ = hex[*p >> 4 & 15];
				*q = hex[*p & 15];
			}
		}
	}

	*q = 0;

	return buf;
}
//---------------------------------------------------------------------------
int gg_http_hash(const char *format, ...)
{
	unsigned int a, c;
	va_list ap;
	int b = -1, i;
    unsigned int j;
    
	va_start(ap, format);

	if (!format)
		return 0;

	for (j = 0; j < strlen(format); j++) {
		unsigned char *arg, buf[16];

		if (format[j] == 'u') {
			snprintf(buf, sizeof(buf), "%d", va_arg(ap, int));
			arg = buf;
		} else {
			if (!(arg = va_arg(ap, unsigned char*)))
				arg = "";
		}

		i = 0;
		while ((c = (int) arg[i++]) != 0) {
			a = (c ^ b) + (c << 8);
			b = (a >> 24) | (a << 8);
		}
	}

	return (b < 0 ? -b : b);
}
//---------------------------------------------------------------------------
int gg_login_hash(char *password, int seed)
{
	unsigned int x, y, z;

	y = seed;

	for (x = 0; *password; password++) {
		x = (x & 0xffffff00) | *password;
		y ^= x;
		y += x;
		x <<= 8;
		y ^= x;
		x <<= 8;
		y -= x;
		x <<= 8;
		y ^= x;

		z = y & 0x1f;
		y = (y << z) | (y >> (32 - z));
	}

	return y;
}
//---------------------------------------------------------------------------
TMenuItem* TGaduThread::StatusToMenuItem (int status)
{
        switch (status)
        {
                case GG_STATUS_AVAIL:
                case GG_STATUS_AVAIL_DESCR:
                        return MainForm->status0;

                case GG_STATUS_BUSY:
                case GG_STATUS_BUSY_DESCR:
                        return MainForm->status1;

                case GG_STATUS_INVISIBLE:
                case GG_STATUS_INVISIBLE_DESCR:
                        return MainForm->status2;

                case GG_STATUS_NOT_AVAIL:
                case GG_STATUS_NOT_AVAIL_DESCR:
                        return MainForm->status3;
        }

        return NULL;
}
//---------------------------------------------------------------------------
void TGaduThread::ChangeStatus (int id, AnsiString desc, bool friends_only, bool force_now)
{
        gg_new_status *s;
        gg_header *h;
        char *buf;

        h = new gg_header;
        s = new gg_new_status;

        if(StatusDescForm)
        {
            asGaduStatus = desc;
            Synchronize(&SetStatusDesc);
        }

        switch (id)
        {
                case 1:
                  if (desc != "") s -> status = GG_STATUS_AVAIL_DESCR; else
                                  s -> status = GG_STATUS_AVAIL;
                break;

                case 2:
                  if (desc != "") s -> status = GG_STATUS_BUSY_DESCR; else
                                  s -> status = GG_STATUS_BUSY;
                break;

                case 3:
                  if (desc != "") s -> status = GG_STATUS_INVISIBLE_DESCR; else
                                  s -> status = GG_STATUS_INVISIBLE;
                break;

                case 4:
                  if (desc != "") s -> status = GG_STATUS_NOT_AVAIL_DESCR; else
                                  s -> status = GG_STATUS_NOT_AVAIL;
                  force_now = true;
                  Terminate();
                break;
        }

        h -> type = GG_NEW_STATUS;
        h -> length = sizeof (int) + ((desc != "") ? desc.Length() : 0);

        if (friends_only) s -> status |= GG_STATUS_FRIENDS_MASK;

        if (force_now)
        {
                try
                {
					if(SS)
					{
						SS -> Write (h, sizeof(gg_header));
						SS -> Write (s, sizeof(int));
						if (desc != "") SS -> Write (desc.c_str(), desc.Length());
                    }
                }
                catch (Exception *E) {}

                delete h;
                delete s;
        } else
        {
                WriteQueue->AddObject (IntToStr(sizeof (gg_header)), (TObject*)h);
                WriteQueue->AddObject (IntToStr(sizeof (int)), (TObject*)s);
                if (desc != "")
                {
                        buf = strdup (desc.c_str());
                        WriteQueue->AddObject (desc.Length(), (TObject*)buf);
                }
        }
}
//---------------------------------------------------------------------------
void TGaduThread::SendNotifyList()
{
        gg_header *h;
		gg_notify *n;
		int iGGcontacts = 0;
		//policz kontakty na licie ktre posiadaj GG UIN
		for (int i = 0; i < MainForm -> List -> Items -> Count; i++)
			if(StrToIntDef(MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0], -1) != -1)    //pusty UIN, np. przy imporcie z AQQ
				iGGcontacts++;

		if (iGGcontacts == 0) {
				h = new gg_header;
				h -> type = GG_LIST_EMPTY;
				h -> length = 0;
				WriteQueue -> AddObject (IntToStr(sizeof (gg_header)), (TObject*)h);
                LogMessage ("GG_LIST_EMPTY sent: no contacts");
        } else
        {
				/** \todo lista moe nie by pusta ale moe by bez UIN */
				h = new gg_header;
                h -> type = GG_NOTIFY_LAST;	///< \todo tu jest ograniczenie do 400 kontaktw
                h -> length = GG_NOTIFY_SIZE * iGGcontacts;
                WriteQueue -> AddObject (IntToStr (sizeof (gg_header)), (TObject*)h);
				for (int i = 0; i < MainForm -> List -> Items -> Count; i++)
				{
						if(StrToIntDef(MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0], -1) != -1)    //pusty UIN, np. przy imporcie z AQQ
						{
							n = new gg_notify;
							n -> uin = StrToInt (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0]);
                            n -> type = StrToInt (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[7]);
                            WriteQueue -> AddObject (IntToStr (GG_NOTIFY_SIZE), (TObject*)n);
                            LogMessage ("[Queue] GG_NOTIFY, uin: " + IntToStr(n->uin) + ", type: " + IntToStr(n -> type));
                        }
                }
        }
}
//---------------------------------------------------------------------------
void TGaduThread::ExecuteWriteQueue()
{
        // LogMessage ("Write Queue Size: " + IntToStr(WriteQueue->Count));

        int count, write;
        BYTE *ptr;

        for (int i = 0; i < WriteQueue -> Count; i++)
        {
                ptr = (BYTE*)WriteQueue -> Objects[i];
                count = StrToInt (WriteQueue -> Strings[i]);
                write = 1;
                while (count > 0 && write > 0)
                        try
                        {
                                write = SS -> Write (ptr, count);
                                count -= write;
                                ptr += write;
                        }
                        catch (Exception *E) {}

                if (count == 0)
                {
                        err_count = 0;
                        LogMessage ("[SEND] Size: " + WriteQueue -> Strings[i]);
                }
                else
                {
                        err_count++;
                        LogMessage ("[SEND] Incomplete. Sent " + IntToStr((StrToInt(WriteQueue -> Strings[i]) - count)) + " out of " + WriteQueue -> Strings[i]);
                }

				/** \todo CG ostrzega, nieza sztuczka z uyciem free */
                free ((void*) WriteQueue -> Objects[i]);
				WriteQueue -> Strings[i] = "";
                LastWrite = GetTickCount();
        }

        for (int i = WriteQueue -> Count - 1; i >= 0; i--)
                if (WriteQueue -> Strings[i] == "") WriteQueue -> Delete(i);
}
//---------------------------------------------------------------------------
void TGaduThread::AddNotify (int uin, char type)
{
        gg_header *h;
        gg_add_notify *a;

        h = new gg_header;
        a = new gg_add_notify;
        h -> type = GG_ADD_NOTIFY;
        h -> length = GG_ADD_NOTIFY_SIZE;
        a -> uin = uin;
        a -> type = type;

        WriteQueue -> AddObject (IntToStr (sizeof (gg_header)), (TObject*)h);
        WriteQueue -> AddObject (IntToStr (GG_ADD_NOTIFY_SIZE), (TObject*)a);

        LogMessage ("[Add notify] UIN: " + IntToStr (a -> uin) + " Type: " + IntToStr(a -> type));
}
//---------------------------------------------------------------------------
void TGaduThread::RemoveNotify (int uin, char type)
{
        gg_header *h;
        gg_remove_notify *r;

        h = new gg_header;
        r = new gg_remove_notify;
        h -> type = GG_REMOVE_NOTIFY;
        h -> length = GG_REMOVE_NOTIFY_SIZE;
        r -> uin = uin;
        r -> type = type;

        WriteQueue -> AddObject (IntToStr (sizeof (gg_header)), (TObject*)h);
        WriteQueue -> AddObject (IntToStr (GG_REMOVE_NOTIFY_SIZE), (TObject*)r);

        LogMessage ("[Remove notify] UIN: " + IntToStr (r -> uin));
}

//---------------------------------------------------------------------------
/** danie operacji dotyczcej kontaktw przechowywanych na serwerze */
void TGaduThread::UserlistRequest (char type)
{
	gg_header *h;


	if(type == GG_USERLIST_GET)
	{
		h = new gg_header;
		gg_userlist_request *a;
		a = new gg_userlist_request;
		h -> type = GG_USERLIST_REQUEST;
		h -> length = GG_USERLIST_REQUEST_SIZE;
		a -> type = type;
		WriteQueue -> AddObject (IntToStr (sizeof (gg_header)), (TObject*)h);
		WriteQueue -> AddObject (IntToStr (GG_ADD_NOTIFY_SIZE), (TObject*)a);
	}
	else if(type == GG_USERLIST_PUT || type == GG_USERLIST_PUT_MORE)
	{
		if(!ContactExportBuffer.Length())
			return;
		h = new gg_header;
		int request_size = MIN(ContactExportBuffer.Length(), 2047);	//2047 poniewa bajt zerowy zarezerwowany na typ
		gg_userlist_request *a;
		a = new gg_userlist_request;
		h -> type = GG_USERLIST_REQUEST;
		h -> length = GG_USERLIST_REQUEST_SIZE + request_size;
		a -> type = type;
		AnsiString req = ContactExportBuffer.SubString(1, request_size);
		ContactExportBuffer = ContactExportBuffer.SubString(request_size, ContactExportBuffer.Length()-request_size);
		WriteQueue -> AddObject (IntToStr (sizeof (gg_header)), (TObject*)h);
		WriteQueue -> AddObject (IntToStr (GG_USERLIST_REQUEST_SIZE), (TObject*)a);
		WriteQueue -> AddObject (IntToStr (req.Length()), (TObject*)(req.c_str()));

	}
	LogMessage ("[UserlistRequest] type = " + IntToStr(type));
}
//---------------------------------------------------------------------------
/** \brief Wywietlenie przychodzcej wiadomoci tekstowej */
void __fastcall TGaduThread::ShowMsg()
{
        TTalkForm *T;
        bool found = false;
        for (int i = 0; i < MainForm->ComponentCount; i++)
                if (MainForm -> Components[i] -> Tag == sender)
                {
                        T = (TTalkForm*) MainForm -> Components[i];
                        T -> Visible = true;
                        found = true;
                }

        if (!found)
        {
                AnsiString display = IntToStr(sender);
                for (int i = 0; i < MainForm -> List -> Items -> Count; i++)
                {
					if (StrToIntDef (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0], -1) == sender)
						if (StrToInt (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[7]) == 4) return; else display = MainForm -> List -> Items -> Item[i] -> Caption;
				}
                T = new TTalkForm (MainForm);
                T -> Caption = "Rozmowa z " + display;
                T -> Tag = sender;
                T -> display = display;
                T -> WindowState = wsMinimized;
                T -> Show();
        }

        T -> Chat -> SelStart = T -> Chat -> Lines -> Text.Length();
        T -> Chat -> SelAttributes -> Size = 8;
        T -> Chat -> SelAttributes -> Style = TFontStyles() << fsBold;
        T -> Chat -> SelAttributes -> Color = clBlack;
        T -> Chat -> Paragraph -> FirstIndent = 0;

        int current_time = time(NULL);
        char buf [30];
        tm *t;
        if (current_time - msg_time <= 300)
        {
                t = localtime ((long*)&current_time);
                strftime (buf, sizeof (buf), "%H:%M:%S", t);
        } else
        {
                t = localtime ((long*)&msg_time);
                strftime (buf, sizeof (buf), "%d/%m, %H:%M:%S", t);
        }
        T -> Chat -> Lines -> Add (T -> display + ": (" + buf + ") ");

        T -> Chat -> SelAttributes -> Size = 10;
        T -> Chat -> SelAttributes -> Style = TFontStyles();
        T -> Chat -> SelAttributes -> Color = clRed;
        T -> Chat -> Paragraph -> FirstIndent = 10;
        T -> Chat -> Lines -> Add (message);

        SendMessage (T -> Chat -> Handle, WM_VSCROLL, SB_BOTTOM, NULL);

        if (!DirectoryExists (T -> log_file + "\\" + IntToStr(T -> Tag))) CreateDir (T -> log_file + "\\" + IntToStr (T -> Tag));
        T -> Chat -> Lines -> SaveToFile (T -> log_file + "\\" + IntToStr(T -> Tag) + "\\" + IntToStr(T -> timestamp) + ".rtf");

        FLASHWINFO f = {sizeof (f), T -> Handle, FLASHW_TRAY | FLASHW_TIMERNOFG, 0, 0};
        if (GetForegroundWindow() != T -> Handle)
        {
                FlashWindowEx (&f);
				if (SettingsForm -> ListView1 -> Items -> Item[4] -> SubItems -> Strings[0] != "0") sndPlaySound (SettingsForm -> ListView1 -> Items -> Item[4] -> SubItems -> Strings[0].c_str(), SND_ASYNC);
        }
}
//---------------------------------------------------------------------------
void TGaduThread::ReceiveMsg (BYTE *packet, INT size)
{
        gg_recv_msg m;
        BYTE *ptr = packet;

        memcpy (&m, ptr, sizeof(m));
        ptr += sizeof (m);

//        LogMessage ("[MSG] Class: " + IntToStr (m.cls));

        // pakiet typu poczenie bezporednie
        if (m.cls == GG_CLASS_CTCP && TransferForm -> CheckBox1 -> Checked)
        {
                TListItem *I = MainForm -> FindUIN (m.sender);
                if (I != NULL)
                {
                        TTransferThread *T = new TTransferThread (true);
                        T -> type = TT_ACTIVE_GET_FILE;
                        T -> d_addr = I -> SubItems -> Strings[8];
                        T -> d_port = StrToInt (I -> SubItems -> Strings[9]);
                        T -> d_uin = m.sender;
                        T -> Resume();
                }
        } else
        if (m.cls != GG_CLASS_CTCP)
        {
                LogMessage (IntToStr(m.sender) + ": " + (char*)ptr);
                sender = m.sender;
                message = (char*)ptr;
                msg_time = m.time;
				Synchronize (&ShowMsg);
        }

        free (packet);
}
//---------------------------------------------------------------------------
void TGaduThread::NotifyReply (BYTE *packet, INT size)
{
		gg_notify_reply n;
		AnsiString msg = "";
		int time;
		BYTE *ptr = packet;

		LogMessage ("[NR] Size: " + IntToStr(size) + " SizeOf(n): " + IntToStr (GG_NOTIFY_REPLY_SIZE));

		memcpy (&n, ptr, GG_NOTIFY_REPLY_SIZE);
		ptr += GG_NOTIFY_REPLY_SIZE;

		if (n.status == GG_STATUS_BUSY_DESCR || n.status == GG_STATUS_AVAIL_DESCR || n.status == GG_STATUS_NOT_AVAIL_DESCR)
		{
				int count;
				char c;

                count = size - GG_NOTIFY_REPLY_SIZE;
				while (count > 0)
                {
                        c = ((char*)ptr)[0];
						ptr++;
                        count--;
                        if (c != 0) msg += c; else break;
				}

                if (count == sizeof (int)) memcpy (&time, ptr, count);

                NotifyReply2 (n.uin, n.status, msg, n.remote_ip, n.remote_port);
        } else
		{
                int count;

				count = size / GG_NOTIFY_REPLY_SIZE;

                NotifyReply2 (n.uin, n.status, "", n.remote_ip, n.remote_port);
				for (int i = 1; i < count; i++)
				{
						memcpy (&n, ptr, GG_NOTIFY_REPLY_SIZE);
						ptr += GG_NOTIFY_REPLY_SIZE;
						NotifyReply2 (n.uin, n.status, "", n.remote_ip, n.remote_port);
				}
		}

		free (packet);
}
//---------------------------------------------------------------------------

/** Wysyany gdy osoba ma mnie na licie ? */
void TGaduThread::NotifyReply60 (BYTE *packet, INT size)
{
		gg_notify_reply60 n;
		AnsiString msg = "";
		int time;
		BYTE *ptr = packet;

		LogMessage ("[NR60] Size: " + IntToStr(size) + " SizeOf(n): " + IntToStr (GG_NOTIFY_REPLY_SIZE60));
		while(size >= GG_NOTIFY_REPLY_SIZE60)
		{
			memcpy (&n, ptr, GG_NOTIFY_REPLY_SIZE60);
			ptr += GG_NOTIFY_REPLY_SIZE60;
			size -= GG_NOTIFY_REPLY_SIZE60;
			n.uin &= 0x00FFFFFF;	//zamaskowanie najstarszego bajtu zawierajcego od wer. 6.0 flagi specjalne (np. GG_UINFLAG_VOICE)
			if (n.status == GG_STATUS_BUSY_DESCR || n.status == GG_STATUS_AVAIL_DESCR || n.status == GG_STATUS_NOT_AVAIL_DESCR)
			{
					//status wraz z opisem
					int count;
					char c;

					count = (int)(*ptr);
					ptr++;	//ptr+1: pominicie pola description_size
                    size -= (count+1);
					while (count > 0)
					{
							c = ((char*)(ptr))[0];
							ptr++;
							count--;
							if (c != 0) msg += c; else break;
					}
					/** \todo o co chodzi? */
					if (count == sizeof (int)) memcpy (&time, ptr, count);
					/** \todo nierozwizana zagadka timestampa */
					//ptr += 4;	//pominicie pola time
					NotifyReply2 (n.uin, n.status, msg, n.remote_ip, n.remote_port);
					msg = "";
			} else
			{
					NotifyReply2 (n.uin, n.status, "", n.remote_ip, n.remote_port);
			}
		}
		free (packet);
}
//---------------------------------------------------------------------------



/** Obsuga odpowiedzi na status uytkownika (NOTIFY) */
void TGaduThread::NotifyReply2 (int uin, int status, AnsiString desc, int ip, int port)
{
	LogMessage ("[Status] UIN: " + IntToStr(uin) + " Status:" + IntToStr(status) + " Desc: " + desc);

	for (int i = 0; i < MainForm -> List -> Items -> Count; i++)
			if (StrToIntDef (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0], -1) == uin)
			{
					// Inwigilator taki maly ;)
					int img_index = StatusToImageIndex (status);
					if (img_index == 0) img_index = 3;
					MainForm -> List -> Items -> Item[i] -> ImageIndex = img_index;
					MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[1] = desc;
					//---- P2P ----
					in_addr in;
					in.S_un.S_addr = ip;
                    char *buf = inet_ntoa (in);

					MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[8] = buf;
					//free (buf);
					MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[9] = IntToStr(port);
					//-------------
					if (desc != "") MainForm -> List -> Items -> Item[i] -> StateIndex = 0; else
									MainForm -> List -> Items -> Item[i] -> StateIndex = -1;
			}

	LogActivity (uin, status, desc);

	MainForm -> List -> CustomSort (NULL, 0);
}
//---------------------------------------------------------------------------


void TGaduThread::StatusUpdate (BYTE *packet, INT size)
{
		gg_status s;
		int time;
		unsigned int count;
        AnsiString msg;
		char c;
        BYTE *ptr = packet;

		LogMessage ("[SU] Size: " + IntToStr(size) + " Sizeof(S): " + IntToStr(sizeof(s)));

        memcpy (&s, ptr, sizeof(s));
        ptr += sizeof(s);

        count = 0;
        c = 1;
        msg = "";
		if (size > sizeof (s))
        {
                while (c != 0 && count < size - sizeof(s))
				{
                        c = ((char*)ptr)[0];
                        ptr++;
						msg = msg + c;
						count++;
				}

				int left = size - sizeof(s) - count;
				if (left > 0) memcpy (&time, ptr, left);
		} else LogMessage ("[SU] Neither time nor desc");

		NotifyReply2 (s.uin, s.status, msg);

		free (packet);
}

//---------------------------------------------------------------------------

void TGaduThread::StatusUpdate60 (BYTE *packet, INT size)
{
		gg_status60 s;
		int time;
		unsigned int count;
        AnsiString msg;
		char c;
        BYTE *ptr = packet;

		LogMessage ("[SU60] Size: " + IntToStr(size) + " Sizeof(S): " + IntToStr(sizeof(s)));

        memcpy (&s, ptr, sizeof(s));
        s.uin &= 0x00FFFFFF;	//pozbd si flag przy najstarszym bajcie
		ptr += sizeof(s);

        count = 0;
        c = 1;
        msg = "";
		if (size > sizeof (s))
        {
                while (c != 0 && count < size - sizeof(s))
				{
                        c = ((char*)ptr)[0];
                        ptr++;
						msg = msg + c;
						count++;
				}

				int left = size - sizeof(s) - count;
				if (left > 0) memcpy (&time, ptr, left);
		} else LogMessage ("[SU60] Neither time nor desc");

		NotifyReply2 (s.uin, s.status, msg, s.remote_ip, s.remote_port);

		free (packet);
}

//---------------------------------------------------------------------------
void TGaduThread::SendMsg (int UIN, AnsiString msg)
{
        gg_header *h;
        gg_send_msg *m;
        char *text;

        h = new gg_header;
        m = new gg_send_msg;

        h -> type = GG_SEND_MSG;
        h -> length = sizeof (gg_send_msg) + msg.Length() + 1;

        m -> recipient = UIN;
        m -> seq = rand();
        m -> cls = GG_CLASS_CHAT;

        text = strdup (msg.c_str());

        WriteQueue -> AddObject (IntToStr(sizeof(gg_header)), (TObject*)h);
        WriteQueue -> AddObject (IntToStr(sizeof(gg_send_msg)), (TObject*)m);
        WriteQueue -> AddObject (msg.Length() + 1, (TObject*)text);

        LogMessage ("[SendMsg] UIN: " + IntToStr(UIN) + " Msg: " + text + " Length: " + IntToStr (msg.Length()) + " Seq: " + IntToStr (m -> seq));
}
//---------------------------------------------------------------------------
void TGaduThread::MessageAck (BYTE *packet)
{
        gg_send_msg_ack a;

        memcpy (&a, packet, sizeof(a));
        LogMessage ("[ACK] UIN: " + IntToStr(a.recipient) + " SEQ: " + IntToStr(a.seq) + " Status: " + IntToStr(a.status));

        free (packet);
}
//---------------------------------------------------------------------------
void TGaduThread::PubdirRequest(int type, int seq, AnsiString request)
{
        char *req;
        gg_header *h;
        gg_pubdir50 *pd;

        req = strdup (request.c_str());
        for (int i = 0; i < request.Length(); i++)
                if (req[i] == '.') req[i] = 0;

        h = new gg_header;
        pd = new gg_pubdir50;

        h -> type = GG_PUBDIR50_REQUEST;
        h -> length = GG_PUBDIR50_SIZE + request.Length();

        LogMessage ("[PUBDIR50] Header: " + IntToStr(h -> length) + " Request: " + request + " Seq: " + IntToStr(seq));

        pd -> type = type;
        pd -> seq = seq;

        WriteQueue -> AddObject (IntToStr(sizeof (gg_header)), (TObject*)h);
        WriteQueue -> AddObject (IntToStr(GG_PUBDIR50_SIZE), (TObject*)pd);
		WriteQueue -> AddObject (IntToStr(request.Length()), (TObject*)req);
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread::ShowSearchResults()
{
        TListItem *I;
        int k = 0;
        AnsiString key = "", value = "";
        bool next = false;

        SearchForm -> ListView2 -> Items -> Clear();

        I = SearchForm -> ListView2 -> Items -> Add();
        for (int i = 0; i < 4; i++) I -> SubItems -> Add("");

        for (int i = 1; i <= Pubdir50_Reply.Length(); i++)
        {
                if (Pubdir50_Reply[i] == 13)
                {
                        k++;
                        if (i < Pubdir50_Reply.Length() && Pubdir50_Reply[i+1] == 13) next = true;
                }
                else
                if (k % 2 == 0) key += Pubdir50_Reply[i]; else
                                value += Pubdir50_Reply[i];

                if (k >= 2)
                {
                        //ShowMessage (key + ": " + value);
                        if (key == "FmNumber") I -> Caption = value; else
                        if (key == "firstname") I -> SubItems -> Strings[0] = value; else
                        if (key == "nickname") I -> SubItems -> Strings[1] = value; else
                        if (key == "birthyear") I -> SubItems -> Strings[2] = value; else
                        if (key == "city") I -> SubItems -> Strings[3] = value; else
                        if (key == "FmStatus") I -> ImageIndex = StatusToImageIndex (StrToInt(value)); else
                        if (key == "nextstart") { I -> Caption = value; I -> SubItems -> Strings[0] = "<-- nextstart"; SearchForm -> fmstart -> Text = value; }
                        key = "";
                        value = "";
                        k = 0;
                }

                if (next)
                {
                        next = false;
                        i++;
                        I = SearchForm -> ListView2 -> Items -> Add();
                        for (int i = 0; i < 4; i++) I -> SubItems -> Add("");
                }
        }
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread :: ShowReadResults()
{
        AnsiString key = "", value = "";
        int k = 0;

        for (int i = 1; i <= Pubdir50_Reply.Length(); i++)
        {
                if (Pubdir50_Reply[i] == 13)
                {
                        k++;
                        if (k >= 2)
                        {
                                for (int j = 0; j < MeForm -> ComponentCount; j++)
                                        if (key == MeForm -> Components[j] -> Name)
                                                if (MeForm -> Components[j] -> Tag == 101) ((TEdit*)MeForm -> Components[j]) -> Text = value; else
                                                if (MeForm -> Components[j] -> Tag == 102) ((TComboBox*)MeForm -> Components[j]) -> ItemIndex = StrToInt (value);

                                key = "";
                                value = "";
                                k = 0;
                        }
                } else
                if (k % 2 == 0) key += Pubdir50_Reply[i]; else
                                value += Pubdir50_Reply[i];
        }

}
//---------------------------------------------------------------------------
void __fastcall TGaduThread :: WriteSuccessful()
{
		MessageBox (MeForm -> Handle, "Dane zostay zapisane w katalogu.", "Informacja", MB_OK | MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread :: ImportSuccessful()
{
		MessageBox (MainForm -> Handle, "Lista kontaktw zostaa zaimportowana z serwera.", MainForm->Caption.c_str(), MB_OK | MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread :: ExportSuccessful()
{
		MessageBox (MainForm -> Handle, "Lista kontaktw zostaa wyeksportowana na serwer.", MainForm->Caption.c_str(), MB_OK | MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------
void TGaduThread :: PubdirReply (BYTE *packet, INT size)
{
        gg_pubdir50_reply r;
        AnsiString reply = "";
        BYTE *ptr = packet;

        LogMessage ("[PUBDIR50] Size: " + IntToStr(size) + " SizeOf(r): " + IntToStr (GG_PUBDIR50_REPLY_SIZE));

        memcpy (&r, ptr, GG_PUBDIR50_REPLY_SIZE);
        ptr += GG_PUBDIR50_REPLY_SIZE;

		if (r.type == GG_PUBDIR50_WRITE) Synchronize(&WriteSuccessful);

        if (size > GG_PUBDIR50_REPLY_SIZE)
        {
                for (int i = 0; i < size - GG_PUBDIR50_REPLY_SIZE; i++)
                       if (ptr[i] == 0) ptr[i] = 13;
                reply += (AnsiString)((char*)ptr);

                LogMessage (((AnsiString)"[PUBDIR50] Type: " + IntToStr(r.type) + " Seq: " + IntToStr(r.seq) + " Reply: ") + reply);

                Pubdir50_Reply = reply;

                switch (r.type)
                {
                        case GG_PUBDIR50_SEARCH_REPLY:
								Synchronize (&ShowSearchResults);
								break;

						case GG_PUBDIR50_SEARCH:
								Synchronize (&ShowReadResults);
								break;
				}
		}

		free (packet);
}
//---------------------------------------------------------------------------

void TGaduThread::UserListReply (BYTE *packet, INT size)
{
	gg_userlist_reply r;
	AnsiString reply = "";
	BYTE *ptr = packet;

	LogMessage ("[UserList Reply] Size: " + IntToStr(size) + " Sizeof(S): " + IntToStr(sizeof(r)));
	if(size < sizeof(r))
		return;
	memcpy (&r, ptr, sizeof(r));
	ptr += sizeof(r);

	switch(r.type)
	{
		case GG_USERLIST_PUT_REPLY:         /* pocztek eksportu listy */
		case GG_USERLIST_PUT_MORE_REPLY:    /* kontynuacja */
			if(ContactExportBuffer.Length())
				UserlistRequest(GG_USERLIST_PUT_MORE);
			else
			{
				LogMessage ("[UserlistRequest] export finished");
				Synchronize(&ExportSuccessful);
			}
			break;
		case GG_USERLIST_GET_MORE_REPLY:    /* pocztek (lub rodek) importu listy */
			if(size > sizeof(r))
				ContactListBuffer += (char*)ptr;
			break;
		case GG_USERLIST_GET_REPLY: 		   /* ostatnia cz importu */
			if(size > sizeof(r))
				ContactListBuffer += (char*)ptr;

			TStringList *slist = new TStringList();
			slist->Text = ContactListBuffer;
			ContactForm->ReadContactsFromList(slist);
			delete slist;
			Synchronize(&ImportSuccessful);
			//SendNotifyList(); // powoduje zerwanie poczenia z serwerem
			break;
	}

	free (packet);

}
//---------------------------------------------------------------------------


int TGaduThread::StatusToImageIndex (int status)
{
		switch (status)
		{
				case GG_STATUS_NOT_AVAIL:
				case GG_STATUS_NOT_AVAIL_DESCR:
						return 0;

                case GG_STATUS_AVAIL:
                case GG_STATUS_AVAIL_DESCR:
                        return 1;

                case GG_STATUS_BUSY:
                case GG_STATUS_BUSY_DESCR:
                        return 2;
        }

        return -1;
}
//---------------------------------------------------------------------------
int gg_fix32(int x)
{
	return (int)
		(((x & (int) 0x000000ffU) << 24) |
                 ((x & (int) 0x0000ff00U) << 8) |
                 ((x & (int) 0x00ff0000U) >> 8) |
                 ((x & (int) 0xff000000U) >> 24));
}

short gg_fix16(short x)
{
	return (short)
		(((x & (short) 0x00ffU) << 8) |
                 ((x & (short) 0xff00U) >> 8));
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread::Execute()
{
        dead = false;

        for (int i = 0; i < WriteQueue ->Count; i++)
                delete ((void*)WriteQueue -> Objects[i]);
        WriteQueue -> Clear();

        LastWrite = GetTickCount();

		CHAR buf [256];
        AnsiString addr, buf2 = "";
        int count;

again:

        LogMessage ("Attempting to get GG server IP...\nConnecting to appmsg.gadu-gadu.pl...");

        SetStatus ("Lcz si...");

        S = new TClientSocket (NULL);

        S -> Host = "appmsg.gadu-gadu.pl";
        S -> Port = 80;
        S -> ClientType = ctBlocking;
        try
        {
                S -> Open();
        }
        catch (Exception *E)
        {
                SetStatus ("URL startowy nie dziala");
                Terminate();
                delete S;
                dead = true;
                return;
        }

        SS = new TWinSocketStream (S -> Socket, 60000); //timeout = 60s
        // cig jest zgodny z wysyanym przez AQQ (stan na 2007.05.17), pozostawiem
        // jednak UIN = 101 zamiast prawdziwego identyfikatora uytkownika jak wysya to AQQ
        S -> Socket -> SendText ("GET /appsvc/appmsg4.asp?fmnumber=101&version=6%2c+0%2c+0%2c+150&lastmsg=0 HTTP/1.0\r\n"
            "Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, */*\r\n"
            "User-Agent: Mozilla/4.0 (compatible; ICS)\r\n"
            "Host: appmsg.gadu-gadu.pl\r\n\r\n");

        while ( (count = SS -> Read (buf, 256)) > 0 ) buf2 += ((AnsiString)buf).SubString (1, count);
        LogMessage (buf2);

        try
        {
                AnsiString ad, ad_nr;
                int i;

                i = buf2.Pos ("\r\n\r\n");
                ad = buf2.SubString (i + 4, buf2.Length() - i - 4);
                ad_nr = ad.SubString (1, ad.Pos (" ")-1);
                i = ad.Pos ("\n");
                ad = ad.SubString (i + 1, ad.Length() - i - 1);
                // ShowMessage (ad_nr);

                if (ad_nr != "0" && ad_nr != "@")
                {
                        SettingsForm -> ListView1 -> Items -> Item[14] -> SubItems -> Strings[0] = ad_nr;
                        if (SettingsForm -> ListView1 -> Items -> Item[13] -> SubItems -> Strings[0] == "1") MessageBox (MainForm -> Handle, ad.c_str(), "Komunikat systemowy", MB_OK | MB_ICONINFORMATION);
                }
        } catch (Exception *E) {}

        CHAR *tmp;

        tmp = strtok (buf2.c_str(), " ");
        while (tmp != NULL) {
                if (StrPos (tmp, "8074") != NULL) {
                        AnsiString t = tmp;

                        addr = t.SubString (1, t.Pos(":")-1);
                        LogMessage ("Our host is: " + addr);
                }
                tmp = strtok (NULL, " ");
        }
        free (tmp);

        delete SS;
        S ->Close();
        delete S;

        S = new TClientSocket (NULL);

        S -> ClientType = ctBlocking;
        S -> Address = addr;
        S -> Port = 8074;

        try
        {
                S -> Open();
        }
        catch (Exception *E)
        {
                SetStatus ("Serwer nie dziala");
                Terminate();
                delete S;
                dead = true;
                return;
        }

        SS = new TWinSocketStream (S -> Socket, 10000);

        gg_header h;
        gg_welcome w;

        SS -> Read (&h, sizeof (h));
        SS -> Read (&w, sizeof (w));

        LogMessage ("Received GG_WELCOME, seed: "+IntToStr(w.seed));

		gg_login60 l;

		h.type = GG_LOGIN60;
		h.length = GG_LOGIN_SIZE60;

		memset (&l, 0, sizeof(l));
		l.uin = StrToInt(SettingsForm->ListView1->Items->Item[0]->SubItems->Strings[0]);
		l.hash = gg_login_hash (SettingsForm->ListView1->Items->Item[1]->SubItems->Strings[0].c_str(), w.seed);
		l.status = StrToInt(SettingsForm->ListView1->Items->Item[2]->SubItems->Strings[0]);
		l.version = 0x20;	//klient podaje si za wersj 6.0
		l.unknown1 = 0x00;
		l.unknown2 = 0xBE;
		l.image_size = 0;

        //---- P2P ----
        if (TransferForm -> CheckBox1 -> Checked)
		{
				if (TransferForm -> CheckBox3 -> Checked)
				{
						l.local_ip = inet_addr (TransferForm -> Edit3 -> Text.c_str());
						l.local_port = StrToInt (TransferForm -> Edit2 -> Text);
				} else
				{
						hostent *he;
						gethostname (buf, sizeof(buf));
						he = gethostbyname (buf);
						l.local_ip = *(DWORD*)*he->h_addr_list;
						if (TransferForm -> CheckBox2 -> Checked)
								l.local_port = 1;
						else
								l.local_port = StrToInt (TransferForm -> Edit1 -> Text);
						free (he);
				}
				/** \todo nie sprawdzone */
				l.external_ip = l.local_ip;
				l.external_port = l.local_port;
		}
		//-------------

        AnsiString desc;
        desc = SettingsForm -> ListView1 -> Items -> Item[9] -> SubItems -> Strings[0];
        if (desc != "")
        {
                h.length += desc.Length();
                StatusDescForm -> Edit1 -> Text = desc;
                MainForm -> Opis1 -> Checked = true;
                l.status += 2;
        }

        if (SettingsForm -> ListView1 -> Items -> Item[11] -> SubItems -> Strings[0] == "1")
        {
                l.status |= GG_STATUS_FRIENDS_MASK;
                MainForm -> fonly -> Checked = true;
        }

		LogMessage ("Sending GG_LOGIN60 / length: " + IntToStr(h.length) + " / desc: " + desc);

		SS -> Write (&h, sizeof (gg_header));
		SS -> Write (&l, GG_LOGIN_SIZE60);
		if (desc != "") SS -> Write (desc.c_str(), desc.Length());

        while (!Terminated)
        {
                if (GetTickCount() - LastWrite >= 290000)
                {
                        gg_header *h_ping = new gg_header;
                        h_ping -> type = GG_PING;
                        h_ping -> length = 0;
                        WriteQueue -> AddObject (IntToStr(sizeof (gg_header)), (TObject*)h_ping);
                        LogMessage ("[PING] ~5 minutes IDLE");
                }

                ExecuteWriteQueue();

                int count = sizeof(gg_header);
                try
                {
                        if (SS -> WaitForData (500))
                        {
                                int read = 1;
                                BYTE *ptr = (BYTE*)&h;
                                while (count > 0 && read > 0)
                                {
                                        read = SS -> Read (ptr, count);
                                        count -= read;
                                        ptr += read;
                                }

                                if (count > 0)
                                {
                                        err_count++;
                                        LogMessage ("[ERR] INCOMPLETE HEADER RECEIVED, err count: " + IntToStr(err_count));
                                } else err_count = 0;
                        }
                }
                catch (Exception *E)
                {
                        LogMessage ("Couldn't read data. Connection has been broken !!!");
                        SetStatus ("Zerwane polaczenie");
                        Terminate();
                }
                
                if (count == 0 && !Terminated)
                {
                        LogMessage ("Header received: " + IntToStr(h.type) + " / " + IntToStr (h.length));
                        // LogMessage ((char*)&h);

                        BYTE *data = NULL, *ptr;
                        int read = 1;

                        count = h.length;
                        if (count < 0 || count > 65535)
                                LogMessage ("[ERR] INVALID PACKET SIZE");
                        else
                        {
                                ptr = data = (BYTE*)malloc (h.length);
                                while (count > 0 && read > 0)
                                        try
                                        {
                                                read = SS -> Read (ptr, count);
                                                count -= read;
                                                ptr += read;
                                        }
                                        catch (Exception *E) {}
                        }

                        if (count != 0)
                        {
                                err_count++;
                                LogMessage ("[ERR] INCOMPLETE PACKET RECEIVED");
                                if (data != NULL) free (data);
                        } else
                        {
                                err_count = 0;

                                switch (h.type)
                                {
                                        case GG_LOGIN_OK:
                                                LogMessage ("GG_LOGIN_OK received - login successful (1715659)");
                                                SetStatus ("Polaczony");
                                                SendNotifyList();
                                                break;

                                        case GG_LOGIN_FAILED:
                                                LogMessage ("GG_LOGIN_FAILED received - login failed");
                                                SetStatus ("Blad przy logowaniu");
                                                Terminate();
                                                break;

                                        case GG_DISCONNECTING:
                                                LogMessage ("We've been disconnected by server. Maybe another session with same UIN ?");
                                                SetStatus ("Rozlaczony przez serwer");
                                                Terminate();
                                                break;

                                        case GG_NOTIFY_REPLY:
                                                LogMessage ("GG_NOTIFY_REPLY received");
                                                NotifyReply (data, h.length);
                                                break;

										case GG_NOTIFY_REPLY60:
												LogMessage ("GG_NOTIFY_REPLY60 received");
												NotifyReply60 (data, h.length);
												break;

										case GG_PONG:
                                                LogMessage ("[PONG]");
                                                break;

                                        case GG_RECV_MSG:
                                                LogMessage ("Header GG_RCV_MSG - someone is talking to us :)");
                                                ReceiveMsg (data, h.length);
                                                break;

                                        case GG_STATUS:
                                                LogMessage ("GG_STATUS received");
												StatusUpdate (data, h.length);
                                                break;

										case GG_STATUS60:
												LogMessage ("GG_STATUS60 received");
												StatusUpdate60 (data, h.length);
												break;

                                        case GG_SEND_MSG_ACK:
                                                LogMessage ("GG_SEND_MSG_ACK received");
                                                MessageAck (data);
                                                break;

                                        case GG_PUBDIR50_REPLY:
                                                LogMessage ("GG_PUBDIR50_REPLY received");
                                                PubdirReply (data, h.length);
												break;

										case GG_USERLIST_REPLY:
												LogMessage ("GG_USERLIST_REPLY received");
												UserListReply (data, h.length);
												break;

                                        default:
                                                free (data);
                                                break;
                                }
                        }
                }

                if (err_count >= 7)
                {
                        SetStatus ("Zerwane polaczenie");
                        Terminate();
                }
        }

        LogMessage ("[GaduThread] Terminated");

        ChangeStatus (4, StatusDescForm -> Edit1 -> Text, false, true);

        S -> Close();
        delete SS;
        delete S;

        // SetStatus ("Rozlaczony");
		Synchronize (&ContactForm -> ResetContactList);

        dead = true;
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread::SetStat()
{
        MainForm -> StatusBar -> Panels -> Items[0] -> Text = status_msg;
}
//---------------------------------------------------------------------------
void __fastcall TGaduThread::SetStatus (AnsiString S)
{
        status_msg = S;
		Synchronize (&SetStat);
}
//---------------------------------------------------------------------------
int activ_uin, activ_status;
AnsiString activ_desc;

void __fastcall TGaduThread::LogActivity (int uin, int status, AnsiString desc)
{
        if (!ActivityForm -> CheckBox1 -> Checked) return;

        activ_uin = uin;
        activ_status = status;
        activ_desc = desc;
		Synchronize (&LogActiv);
}

void __fastcall TGaduThread::LogActiv()
{
        AnsiString s = "---", n = "---";

        switch (activ_status)
        {
                case GG_STATUS_NOT_AVAIL:
                        s = "Niedostepny";
                break;

                case GG_STATUS_NOT_AVAIL_DESCR:
                        s = "Niedostepny +";
                break;

                case GG_STATUS_AVAIL:
                        s = "Dostepny";
                break;

                case GG_STATUS_AVAIL_DESCR:
                        s = "Dostepny +";
                break;

                case GG_STATUS_BUSY:
                        s = "Zaraz wracam";
                break;

                case GG_STATUS_BUSY_DESCR:
                        s = "Zaraz wracam +";
                break;
        }

        TListItem *I = ActivityForm -> ListView -> Items -> Add();
        I -> Caption = DateToStr (Date()) + " " + TimeToStr (Time());
        I -> SubItems -> Add (IntToStr (activ_uin)); // UIN

        for (int i = 0; i < MainForm -> List -> Items -> Count; i++)
			if (StrToIntDef (MainForm -> List -> Items -> Item[i] -> SubItems -> Strings[0], -1) == activ_uin)
            	n = MainForm -> List -> Items -> Item[i] -> Caption;
        /** \todo podejrzane */
		I -> SubItems -> Add (n); // Ksywa
        I -> SubItems -> Add (s); // Status
        I -> SubItems -> Add (activ_desc); // Opis

        if (!ActivityForm -> CheckBox2 -> Checked) return;

        HWND hwnd;

		if (!MyPopupForm -> Visible)
		{
				MyPopupForm -> Text -> Caption = "";
				MyPopupForm -> ClientHeight = 10;
				MyPopupForm -> Timer1 -> Enabled = true;
                hwnd = GetForegroundWindow();
        }

		MyPopupForm -> Timer1 -> Tag = 5;
		MyPopupForm -> Text -> Caption = "[" + Time().FormatString ("hh:nn") + "] " + n + ": " + s + "\n" + MyPopupForm -> Text -> Caption;
		if (MyPopupForm -> ClientHeight < 75) MyPopupForm -> ClientHeight += 13;

        APPBARDATA p;

        p.cbSize = sizeof (APPBARDATA);
        p.hWnd = MainForm -> Handle;
        SHAppBarMessage (ABM_GETTASKBARPOS, &p);

		MyPopupForm -> Left = p.rc.right - MyPopupForm -> Width - 10;
		MyPopupForm -> Top = p.rc.top - MyPopupForm -> Height - 10;

		if (!MyPopupForm -> Visible)
		{
				MyPopupForm -> Show();
				SetForegroundWindow (hwnd);
				SetWindowPos (MyPopupForm -> Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE);
		}
}
//---------------------------------------------------------------------------

